#!/usr/bin/perl
# Converts dependencies from CoNLL-06 format to CSTS (Prague Dependency Treebank 1.0) format.
# Note: this script reads and writes UTF-8 while CSTS should be in ISO 8859-2.
# Tred will override any encoding setting by ISO 8859-2 when reading CSTS.
# As a result, the data must be further converted to FS in order to be displayed in Tred.
# Copyright  2006, 2009 Dan Zeman <zeman@ufal.mff.cuni.cz>
# 2.3.2009 ... adapted to the new CoNLL 2009 format (more columns in new order)
# 25.3.2009 ... added option to convert tags to the PDT schema. DZ Interset libraries are needed.
# License: GNU GPL

sub usage
{
    print STDERR ("Usage: conll2csts.pl [-y year] -l language [-pdttags] < input.conll > output.csts\n");
    print STDERR ("Year is 2006 (default) or 2009.\n");
    print STDERR ("-pdttags means that POS tags and features will be converted into the PDT tagset.\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
use Getopt::Long;
use tagset::common;

# CoNLL format changes over years. Default is 2006 and covers 2007 as well. Another option is 2009.
$year = 2006;
$pdttags = 0;
# Mandatory information: -l ISO-639 language code.
GetOptions('year=s' => \$year, 'language=s' => \$lang, 'pdttags' => \$pdttags);
if($lang eq '')
{
    usage();
    print STDERR ("Missing language (ISO 639-1 code).\n");
    print STDERR ("Codes of some CoNLL languages: ar = Arabic, bg = Bulgarian, ca = Catalan, cs = Czech, da = Danish, de = German, el = Greek, en = English, es = Spanish, eu = Basque, hu = Hungarian, ja = Japanese, nl = Dutch, pt = Portuguese, sl = Slovene, sv = Swedish, tr = Turkish, zh = Chinese\n");
    die;
}
elsif($pdttags)
{
    # We have been asked to convert part of speech tags into the tagset of the Prague Dependency Treebank.
    # Thus we need a driver for the source tagset.
    if($lang eq 'cs')
    {
        if($year==2006)
        {
            $srctagset = 'cs::conll';
        }
        else
        {
            $srctagset = 'cs::conll2009';
        }
    }
    elsif($lang eq 'en')
    {
        if($year==2006)
        {
            $srctagset = 'en::conll';
        }
        else
        {
            $srctagset = 'en::conll2009';
        }
    }
    elsif($lang eq 'de')
    {
        if($year==2006)
        {
            $srctagset = 'de::conll';
        }
        else
        {
            $srctagset = 'de::conll2009';
        }
    }
    else
    {
        $srctagset = 'unknown';
        die("Unknown tagset for language $lang. Cannot convert to PDT tags.\n");
    }
}
print("<csts lang=\"$lang\">\n");
# omit optional <h> (header)
print("<doc file=\"file\" id=\"1\">\n");
# <a> header of document is not optional
print("<a>\n");
# include only mandatory elements of <a>
print("<mod>s\n"); # s = current language; d = old language; o = transcribed speech
print("<txtype>mix\n"); # ima = imaginative, inf = informative, mix = mixed
print("<genre>mix\n"); # mixed or unknown genre
print("<med>net\n"); # b = book, nws = newspaper, j = journal (periodic), scr = screenplay, net = Internet, oc = occasional
print("<lang>$lang\n"); # language of the document
print("<temp>0000\n"); # four-digit year of publication
print("<opus>file\n"); # according to Czech National Corpus definitions, this shall be equal to filename
print("<id>1\n"); # document identification within a file, equals to the id attribute of <doc>
print("</a>\n");
# All sentences will be in one chapter and one paragraph.
print("<c>\n");
print("<p>\n");
$i_sentence = 1;
$new_sentence = 1;
while(<>)
{
    # Blank line signals new sentence.
    # Take several consecutive blank lines as one sentence break (avoid empty sentences).
    if(!$new_sentence && m/^\s*$/)
    {
        $i_sentence++;
        $new_sentence = 1;
    }
    # Any non-blank line is a token.
    else
    {
        # If this is the first word of the sentence, print the sentence start tag.
        if($new_sentence)
        {
            print("<s id=\"$i_sentence\">\n");
            $new_sentence = 0;
        }
        # Strip the line break.
        s/\r?\n$//;
        # Encode XML special characters as entities.
        s/&/&amp;/g;
        s/</&lt;/g;
        s/>/&gt;/g;
        # Get field values and print them.
        my ($id, $form, $lemma, $cpostag, $postag, $feats, $head, $deprel, $plemma, $ppos, $pfeat, $phead, $pdeprel, $fillpred, $pred, @apreds);
        my ($tag, $ptag, $pdttag, $ppdttag);
        if($year == 2009)
        {
            # Warning: the PHEAD and PDEPREL occur in both CoNLL 2009 and 2006 but have totally different meanings!
            # We could call them differently but we do not use their values so far so it does not make a difference.
            ($id, $form, $lemma, $plemma, $postag, $ppos, $feats, $pfeat, $head, $phead, $deprel, $pdeprel, $fillpred, $pred, @apreds) = split(/\s+/, $_);
            $tag = $feats eq '_' ? $postag : "$postag\t$feats";
            $ptag = $pfeat eq '_' ? $ppos : "$ppos\t$pfeat";
            if($pdttags)
            {
                $pdttag = tagset::common::encode('cs::pdt', tagset::common::decode($srctagset, $tag));
                $ppdttag = tagset::common::encode('cs::pdt', tagset::common::decode($srctagset, $ptag));
                print("<f>$form<l>$lemma<t>$pdttag<MDl>$plemma<MDt>$ppdttag<r>$id<g>$head<A>$deprel<MDg>$phead<MDA>$pdeprel\n");
            }
            else
            {
                print("<f>$form<l>$lemma<t>$tag<MDl>$plemma<MDt>$ptag<r>$id<g>$head<A>$deprel<MDg>$phead<MDA>$pdeprel\n");
            }
        }
        else # default year = 2006
        {
            # Note: $pdttag is not part of CoNLL format but can be added by a tag conversion utility.
            # It can combine information from $cpostag, $postag, and $feats.
            ($id, $form, $lemma, $cpostag, $postag, $feats, $head, $deprel, $phead, $pdeprel, $pdttag) = split(/\s+/, $_);
            $tag = "$cpostag\t$postag\t$feats";
            if($pdttags)
            {
                $pdttag = tagset::common::encode('cs::pdt', tagset::common::decode($srctagset, $tag));
                print("<f>$form<l>$lemma<t>$pdttag<r>$id<g>$head<A>$deprel\n");
            }
            else
            {
                print("<f>$form<l>$lemma<t>$tag<r>$id<g>$head<A>$deprel\n");
            }
        }
    }
}
print("</p>\n");
print("</c>\n");
print("</csts>\n");
