#!/usr/bin/perl
# Převede treebank z formátu CSTS do formátu CoNLL.
# Copyright © 2007, 2009 Dan Zeman <zeman@ufal.mff.cuni.cz>
# 4.3.2009 ... adapted to the new CoNLL 2009 format (more columns in new order)
# Licence: GNU GPL

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use Getopt::Long;
use csts1;

# CoNLL format changes over years. Default is 2006 and covers 2007 as well. Another option is 2009.
$year = 2006;
GetOptions('year=s' => \$year);
csts::cist(\&zpracovat_vetu);



#------------------------------------------------------------------------------
# Zpracuje strom.
#------------------------------------------------------------------------------
sub zpracovat_vetu
{
    my $veta = shift;
    # V tomto programu používáme atribut parentord, který se ale defaultně čte z prvku <g>.
    # Protože momentálně program používáme pouze pro výstup DZ Parseru, přepíšeme hodnoty
    # atributu parentord hodnotami z <MDg src="dz">.
    foreach my $slovo (@{$veta})
    {
        $slovo->{parentord} = $slovo->{s}{mddz}[0]{pord};
        $slovo->{afun} = $slovo->{s}{mddz}[0]{stag};
    }
    # Pro každý uzel získat vedle jeho skutečného rodiče ještě jeho projektivního předka.
    # Dělá se to jenom pro CoNLL 2006 a 2007, protože pozdější verze formátu CoNLL úplně změnila význam polí PHEAD a PDEPREL.
    if($year<2009)
    {
        # Pozor, pokud by současná struktura nebyl strom, funkce by se mohla zacyklit, proto nejdříve zkontrolovat strom.
        if(csts::je_strom($veta, 'parentord'))
        {
            zprojektivizovat($veta);
        }
        else
        {
    #        print STDERR (join(' ', map {"$_->{form}:$_->{ord}:$_->{parentord}"} @{$veta}), "\n");
    #        die("Cannot convert cyclic graphs.\n");
        }
    }
    for(my $i = 1; $i<=$#{$veta}; $i++)
    {
        my $slovo = $veta->[$i];
        # Validátor nesnáší prázdné řetězce, musíme je nahradit podtržítky.
        foreach my $klic (qw(ord form lemma tag parentord afun projparentord afun))
        {
            if($slovo->{$klic} eq "")
            {
                $slovo->{$klic} = "_";
            }
        }
        if($year == 2009)
        {
            my ($pos, $feat) = split(/\t/, $slovo->{tag});
            my @cells =
            (
                $slovo->{ord},
                $slovo->{form},
                $slovo->{lemma},
                $slovo->{lemma},
                $pos,
                $pos,
                $feat,
                $feat,
                $slovo->{parentord},
                $slovo->{parentord},
                $slovo->{afun},
                $slovo->{afun},
                '_', # FILLPRED
                '_', # PRED
            );
            # Hodnoty musí být neprázdné. Pokud jsou prázdné, nahradit je podtržítkem.
            map {$_ = '_' if(m/^\s*$/)} @cells;
            print(join("\t", @cells), "\n");
        }
        else
        {
            # Tag v CSTS by měl být sřetězením položek CPOSTAG, POSTAG a FEATS.
            # Pokud není, bude POSTAG a FEATS prázdné, resp. níže se do nich doplní podtržítka.
            my @tag = split(/\t/, $slovo->{tag});
            my @cells =
            (
                $slovo->{ord},
                $slovo->{form},
                $slovo->{lemma},
                $tag[0],
                $tag[1],
                $tag[2],
                $slovo->{parentord},
                $slovo->{afun},
                $slovo->{projparentord},
                $slovo->{afun}
            );
            # Hodnoty musí být neprázdné. Pokud jsou prázdné, nahradit je podtržítkem.
            map {$_ = '_' if(m/^\s*$/)} @cells;
            print(join("\t", @cells), "\n");
        }
    }
    print("\n");
}



#------------------------------------------------------------------------------
# Ke každé neprojektivní závislosti najde projektivní závislost. Používá
# jednoduchý algoritmus. Není-li závislost projektivní, nahradí řídící uzel
# jeho rodičem a toto opakuje, dokud závislost není projektivní. To musí nastat
# nejpozději ve chvíli, kdy se řídícím uzlem stane kořen stromu.
#
# POZOR! Tato funkce předpokládá, že dostala opravdový strom bez cyklů!
# Přítomnost cyklů nekontroluje, ale pokud tam nějaké jsou, nikdy neskončí!
#------------------------------------------------------------------------------
sub zprojektivizovat
{
    # pokud je ridici uzel koren stromu, hrana je projektivni (i kdyby meziuzly meli rodice vne intervalu, zprostredkovane nakonec zavisi na koreni)
    # pokud rodic ridiciho uzlu lezi v intervalu, muzeme rovnou posunout ridici uzel o patro vys
    # pokud ma nektery uzel v intervalu rodice mimo interval, jeste to neznamena, ze nelezi v podstromu ridiciho uzlu
    # proto se od kazdeho takoveho uzlu musime vydat smerem ke koreni
    # potkame-li cestou ridici uzel, vychozi uzel prohlasime za schvaleny
    # nepotkame-li ho, udelame ridici uzel z rodice dosavadniho ridiciho a vratime se na zacatek
    my $veta = shift;
    for(my $i = 1; $i<=$#{$veta}; $i++)
    {
        my $z = $i;
        my $r = $veta->[$i]{parentord};
        # Jestliže v průběhu jednoho průchodu cyklem najdeme narušení projektivity hrany $r-$z,
        # uložíme si to do proměnné $neproj a zaručíme tím, že ještě bude další průchod.
        my $neproj = 1;
        while($neproj)
        {
            $neproj = 0;
            # Zapamatovat si, který z obou uzlů je vlevo a který vpravo.
            my $l;
            my $p;
            if($z<$r)
            {
                $l = $z;
                $p = $r;
            }
            else
            {
                $l = $r;
                $p = $z;
            }
            # Vyskočit z cyklu, jestliže jsme u kořene.
            last if($r==0);
            # Jestliže prarodič leží uvnitř intervalu, udělat z něj nový řídící uzel.
            my $ded = $veta->[$r]{parentord};
            if($ded>$l && $ded<$p)
            {
                $neproj = 1;
                $r = $ded;
                next;
            }
            # Projít uzly mezi řídícím a závislým.
            for(my $j = $l; $j<$p; $j++)
            {
                # Pokusit se dokázat, že tento uzel je potomkem řídícího.
                my $k;
                for($k = $j; $k!=0; $k = $veta->[$k]{parentord})
                {
                    # Jestliže jsme došli k řídícímu uzlu, uzel $j nenarušuje projektivitu a můžeme přejít ke kontrole $j+1.
                    last if($k==$r);
                }
                # Pokud jsme v předcházející smyčce došli až do kořene, uzel $j narušuje projektivitu.
                if($k==0)
                {
                    $neproj = 1;
                    $r = $veta->[$r]{parentord};
                    last;
                }
            }
        } # while($neproj)
        # Nyní máme upravený uzel $r, o kterém platí, že hrana $r-$z je projektivní.
        $veta->[$z]{projparentord} = $r;
    }
}
