#!/usr/bin/perl
# Converts dependencies (CSTS format) to constituents (Penn mrg format).
# (c) 2006 Dan Zeman <zeman@ufal.mff.cuni.cz>
# License: GNU GPL

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use csts;



foreach my $file (@ARGV)
{
    csts::projit_data($file, {kodovani_data => "utf8"}, \&process_sentence);
}
###!!! DEBUG: Number of left round brackets.
if(0)
{
    foreach my $n (sort{$b<=>$a}(keys(%stat)))
    {
        print STDERR ("$n left round brackets in $stat{$n} cases\n");
    }
}



###############################################################################
# SUBROUTINES
###############################################################################



#------------------------------------------------------------------------------
# Processes a sentence. Called back from csts.pm.
#------------------------------------------------------------------------------
sub process_sentence
{
    my $state = shift; # state of the reading (hashref)
    my $anot = shift; # annotations (ref to array of hashes)
    # Build top-down dependency links.
    for(my $i = 1; $i<=$#{$anot}; $i++)
    {
        # Make sure there are no cycles.
        for(my $j = $i; $j!=0; $j = $anot->[$j]{rodic_vzor})
        {
            if($anot->[$j]{"touch$i"})
            {
                print STDERR ("Cycle found in document $state->{dokid}, paragraph $state->{odstid}, sentence $state->{vetid}, beginning node $i, repeatedly visited node $j.\n");
                # Print sentence for easy error spotting.
                for(my $k = 1; $k<=$#{$anot}; $k++)
                {
                    print STDERR ("<word>$k $anot->[$k]{slovo} <parent>$anot->[$k]{rodic_vzor} $anot->[$anot->[$k]{rodic_vzor}]{slovo}");
                    if($k==$i || $k==$j)
                    {
                        print STDERR (" <===");
                    }
                    print STDERR ("\n");
                }
                die;
            }
            $anot->[$j]{"touch$i"} = 1;
        }
        # Create link from parent to child.
        push(@{$anot->[$anot->[$i]{rodic_vzor}]{children}}, $anot->[$i]);
    }
    # Build constituents recursively.
    my $parse = build_constituent($anot->[0]);
    # Add unary initial rewrite.
    $parse =~ s/^\(TOP /(TOP (S /;
    $parse .= ")";
    # Remove spaces before right sentences. Evalb cannot eat them.
    $parse =~ s/\s+\)/)/g;
    # Count left round brackets.
    my $brackets = $parse;
    $brackets =~ s/[^\(]//g;
    my $n = length($brackets);
    $stat{$n}++;
    if($n<=2)
    {
        print STDERR ("Warning: Skipping empty parse:\n$parse\n");
    }
    else
    {
        print("$parse\n");
    }
}



#------------------------------------------------------------------------------
# Brackets and returns a constituent. Recursively calls itself for subconsts.
#------------------------------------------------------------------------------
sub build_constituent
{
    my $word = shift;
    # Sort a list of my children and myself.
    my @subs = @{$word->{children}};
    unless($word->{ord}==0)
    {
        push(@subs, $word);
    }
    @subs = sort{$a->{ord}<=>$b->{ord}}(@subs);
    # Construct constituent label.
    my $label;
    if($word->{ord}==0)
    {
        $label = "TOP";
    }
    else
    {
        $label = uc($word->{znacka});
        if(scalar(@subs)>1)
        {
            $label = substr($label, 0, 3) if(length($label)>3);
            $label .= "P";
            # If the data is tagged using the Penn Treebank tagset, our nonterminals can match Penn Treebank as well.
            if($label =~ m/^NNP?S?P$/)
            {
                $label = "NP";
            }
            elsif($label eq "FWP")
            {
                $label = "NP";
            }
            elsif($label =~ m/^JJ[RS]?P$/)
            {
                $label = "ADJP";
            }
            elsif($label eq "DTP")
            {
                $label = "ADJP";
            }
            elsif($label eq "WDTP")
            {
                $label = "WHNP";
            }
            elsif($label eq "PRPP")
            {
                $label = "NP";
            }
            elsif($label eq "WPP")
            {
                $label = "WHPP";
            }
            elsif($label eq "CDP")
            {
                $label = "QP";
            }
            elsif($label =~ m/^VB[DGN]?P$/)
            {
                $label = "VP";
            }
            elsif($label =~ m/^RB[RS]?P$/)
            {
                $label = "ADVP";
            }
            elsif($label eq "WRBP")
            {
                $label = "WHADVP";
            }
            elsif($label eq "INP")
            {
                $label = "PP";
            }
            # "Particle phrase" appears if the infinitival "to" is tagged as particle and marked head of the clause.
            elsif($label =~ m/^(TOP|RPP|SYMP|:P)$/)
            {
                $label = "SBAR";
            }
            # "Conjunction phrase" appears if the sentence begins with "and".
            elsif($label eq "CCP")
            {
                $label = "CONJP";
            }
            elsif($label eq "UHP")
            {
                $label = "INTJ";
            }
        }
    }
    my $constituent = "($label ";
    # Replace each subconstituent by its visual representation.
    foreach my $s (@subs)
    {
        if($s->{ord}==$word->{ord})
        {
            my $w = $s->{slovo};
            # Escape brackets.
            $w =~ s/\(/-LRB-/g;
            $w =~ s/\)/-RRB-/g;
            if(scalar(@subs)>1)
            {
                my $label = uc($s->{znacka});
                # This subconstituent is the head of the current constituent.
                # Mark it as the head using "=H".
                $s = "($label=H $w )";
            }
            else
            {
                $s = $w;
            }
        }
        else
        {
            $s = build_constituent($s);
        }
    }
    $constituent .= join(" ", @subs);
    $constituent .= " )";
    return $constituent;
}
