#!/usr/bin/perl
# Corrects punctuation tagging according to the gold standard treebank. Eliminates evalb mismatch errors.
# (c) 2007 Dan Zeman <zeman@ufal.mff.cuni.cz>
# License: GNU GPL

sub usage
{
    print STDERR ("Usage: evalbpunc.pl -g gold.penn < parsed.penn > corrected.penn\n");
    print STDERR ("  gold.penn:      gold standard trees and tags\n");
    print STDERR ("  parsed.penn:    parser output to correct punctuation in\n");
    print STDERR ("  corrected.penn: corrected parser output, can be safely evaluated using evalb\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use Getopt::Long;
use penn;

# Get options.
GetOptions
(
    'gold=s' => \$gold,
);
if($gold eq "")
{
    usage();
    die("Missing gold standard file name.\n");
}
# Open the gold standard file.
open(GOLD, $gold) or die("Cannot read from $gold: $!\n");
# Read both files and compare them.
while(<>)
{
    my $parsed = $_;
    my $gold = <GOLD>;
    my @gsym = penn::zsplit($gold);
    my @psym = penn::zsplit($parsed);
    # Check punctuation tags in psym. Warning: the lengths of the two arrays can differ!
    my $j = -1;
    for(my $i = 0; $i<$#psym; $i++)
    {
        # Only left brackets with nonterminals, preceding a terminal.
        my $ptag;
        my $gtag;
        if($psym[$i] =~ m/^\((.*)/ && $psym[$i+1] !~ m/^[\(\)]/)
        {
            $ptag = $1;
            # Find next gold tag.
            while($j<$#gsym)
            {
                $j++;
                if($gsym[$j] =~ m/^\((.*)/ && $gsym[$j+1] !~ m/^[\(\)]/)
                {
                    $gtag = $1;
                    last;
                }
            }
            # Remove function (=H means "this is head of the phrase").
            $gtag =~ s/=.*//;
            $ptag =~ s/=.*//;
            # Make sure @psym and @gsym are well synchronized. The next terminals should match.
            unless($psym[$i+1] eq $gsym[$j+1])
            {
                print STDERR ("Warning: terminal mismatch: gold $gsym[$j+1] != parsed $psym[$i+1].\n");
                print STDERR ("  Gold tree:   $gold\n");
                print STDERR ("  Parsed tree: $parsed\n");
            }
            # If gold is punctuation and parsed is not, copy tag to parsed.
            if(is_punctuation($gtag) && !is_punctuation($ptag))
            {
                $psym[$i] = "($gtag";
            }
            # If parsed is punctuation and gold is not, reset parsed tag to SYM.
            elsif(is_punctuation($ptag) && !is_punctuation($gtag))
            {
                $psym[$i] = "(SYM";
            }
        }
    }
    # Tags have been corrected, rejoin bracketing.
    $parsed = penn::zjoin(@psym);
    print($parsed);
}



#------------------------------------------------------------------------------
# Figures out whether a tag is punctuation.
#------------------------------------------------------------------------------
sub is_punctuation
{
    my $tag = shift;
    my $answer = $tag =~ m/^(-LRB-|-RRB-|\pP+)$/;
    return $answer;
}
