#!/usr/bin/perl
# Normalizuje dánské stromy (např. aby přívlastek visel pod podstatným jménem a ne nad ním).
# Současně nahradí slova morfologickými značkami (podrobnými severogermánskými)
# a značky značkami Penn Treebanku.
# (c) 2006 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use csts1;
use tagset::da::conll;
use tagset::en::penn;
use tagset::sv::svdahybrid;
use tagset::sv::hajic;



csts::cist(\&zpracovat_vetu);



#------------------------------------------------------------------------------
# Zpracuje strom.
#------------------------------------------------------------------------------
sub zpracovat_vetu
{
    my $veta = shift;
    # Dekódovat morfologické značky do sad vlastností.
    foreach my $uzel (@{$veta})
    {
        $uzel->{features} = tagset::da::conll::decode($uzel->{tag});
        $uzel->{tag} = tagset::en::penn::encode($uzel->{features});
        $uzel->{form} = tagset::sv::svdahybrid::encode($uzel->{features});
    }
    # Projít uzly, hledat neobvyklá zavěšení a opravovat je.
    foreach my $uzel (@{$veta})
    {
        my $zmena = 1;
        while($zmena)
        {
            $zmena = 0;
            # Přídavné jméno by mělo viset na podstatném a ne naopak.
            if($uzel->{features}{pos} eq "noun" && $uzel->{parent}{features}{pos} eq "adj")
            {
                my $podstatne_jmeno = $uzel;
                my $pridavne_jmeno = $uzel->{parent};
                pripojit($podstatne_jmeno, $pridavne_jmeno->{parent});
                pripojit($pridavne_jmeno, $podstatne_jmeno);
                $zmena = 1;
            }
            # Přivlastňovací, ukazovací a neurčitá zájmena by měla viset na podstatném jméně a ne naopak
            if($uzel->{features}{pos} eq "noun" &&
               $uzel->{parent}{features}{pos} =~ m/^(det|pron)$/ &&
               $uzel->{parent}{features}{synpos} eq "attr")
            {
                # Prohodit tyto dva uzly ve stromu.
                my $podstatne_jmeno = $uzel;
                my $zajmeno = $uzel->{parent};
                # Posbírat seznam dětí, které se budou stěhovat.
                my @deti_ke_stehovani = grep{1}(@{$zajmeno->{children}});
                foreach my $dite (@deti_ke_stehovani)
                {
                    pripojit($dite, $podstatne_jmeno);
                }
                pripojit($podstatne_jmeno, $zajmeno->{parent});
                pripojit($zajmeno, $podstatne_jmeno);
                $zmena = 1;
            }
            # Jestliže visí podstatné jméno na číslovce, převěsit je.
            # Pokud navíc na číslovce visela přídavná jména, zájmena nebo další podstatná jména, převěsit je také pod podstatné jméno.
            if($uzel->{features}{pos} eq "noun" && $uzel->{parent}{features}{pos} eq "num")
            {
                my $podstatne_jmeno = $uzel;
                my $cislovka = $uzel->{parent};
                # Posbírat seznam dětí číslovky, které se budou stěhovat.
                # Nebudeme je rovnou odpojovat, abychom si nenarušili procházení seznamu.
                my @deti_ke_stehovani = grep{1}(@{$cislovka->{children}});
                # Přestěhovat děti.
                foreach my $dite (@deti_ke_stehovani)
                {
                    pripojit($dite, $podstatne_jmeno);
                }
                # Prohodit podstatné jméno s číslovkou.
                pripojit($podstatne_jmeno, $cislovka->{parent});
                pripojit($cislovka, $podstatne_jmeno);
                $zmena = 1;
            }
            # Jestliže visí podstatné jméno na genitivu, prohodit je.
            if($uzel->{features}{pos} eq "noun" && $uzel->{parent}{features}{case} eq "gen")
            {
                my $genitiv = $uzel->{parent};
                my @deti_ke_stehovani = grep{1}(@{$genitiv->{children}});
                # Přestěhovat děti.
                foreach my $dite (@deti_ke_stehovani)
                {
                    pripojit($dite, $uzel);
                }
                pripojit($uzel, $genitiv->{parent});
                pripojit($genitiv, $uzel);
                $zmena = 1;
            }
            # Jestliže visí přídavné jméno na zájmenu, na kterém nevisí současně podstatné jméno, prohodit je.
            if($uzel->{features}{pos} eq "adj" &&
               $uzel->{parent}{features}{pos} =~ m/^(det|pron)$/ &&
               $uzel->{parent}{features}{synpos} eq "attr" &&
               !grep{$_->{features}{pos} eq "noun"}(@{$uzel->{parent}{children}}))
            {
                my $zajmeno = $uzel->{parent};
                my @deti_ke_stehovani = grep{$_!=$uzel}(@{$zajmeno->{children}});
                foreach my $dite (@deti_ke_stehovani)
                {
                    pripojit($dite, $uzel);
                }
                pripojit($uzel, $zajmeno->{parent});
                pripojit($zajmeno, $uzel);
                $zmena = 1;
            }
            # Koordinace jsou v původním dánském treebanku řešeny následovně:
            # - První člen kooridnace je kořenem kooridnace.
            # - Čárky, vnitřní členy a spojka (např. "og") visí na prvním členu.
            # - Poslední člen koordinace visí na spojce.
            # - Afun prvního členu popisuje funkci celé koordinace.
            # - Spojka má afun="coord", čárky mají afun="pnct".
            # - Další členy mají afun="conj".
            # Chceme, aby byly řešeny jako ve švédském treebanku, tedy takto:
            # - První člen koordinace je kořenem koordinace.
            # - Každý další člen visí na předcházejícím členu.
            # - Čárky a spojka visí na členu, před kterým leží ve větě. Nejhlubší list je tedy spojka (např. "och").
            # Koordinaci najdeme podle analytické funkce druhého člena.
            # U všech afunů musíme porovnávat jen prefix, protože kvůli jiným operacím už by mohly mít přívlastek "(převěšeno)".
            if($uzel->{afun} =~ m/^conj/)
            {
                my $koren;
                # Pokud má rodič tohoto uzlu afun coord, je to spojka a náš uzel je poslední člen koordinace.
                if($uzel->{parent}{afun} =~ m/^coord/)
                {
                    $koren = $uzel->{parent}{parent};
                }
                # Jinak je kořenem koordinace přímo rodič tohoto uzlu.
                else
                {
                    $koren = $uzel->{parent};
                }
                unless($uzel->{_koordinace_zpracovana} || $koren->{_koordinace_zpracovana})
                {
                    # Získat seznam uzlů, které se na koordinaci podílejí: členů, čárek a spojek.
                    my @kooruzly = sort{$a->{ord} <=> $b->{ord}}(grep{$_->{afun} =~ m/^(conj|pnct|coord)/}(@{$koren->{children}}));
                    # Na spojkách ještě visí členy koordinace, ty taky přidat do pole.
                    for(my $i = 0; $i<=$#kooruzly; $i++)
                    {
                        if($kooruzly[$i]{afun} =~ m/^coord/)
                        {
                            foreach my $dite (sort{$a->{ord} <=> $b->{ord}}(@{$kooruzly[$i]{children}}))
                            {
                                if($dite->{afun} =~ m/^conj/)
                                {
                                    push(@kooruzly, $dite);
                                }
                            }
                        }
                    }
                    # Znova seřadit pole koordinačních uzlů podle pořadí ve větě.
                    @kooruzly = sort{$a->{ord} <=> $b->{ord}}(@kooruzly);
                    # Projít koordinační uzly a správně je rozvěsit.
                    # Pamatovat si poslední navštívený člen koordinace.
                    my $predch = $koren;
                    # Projít sourozence a hledat členy koordinace, čárky a spojky.
                    for(my $i = 0; $i<=$#kooruzly; $i++)
                    {
                        # Čárku nebo spojku pověsit na člen koordinace, který je ve větě za ní.
                        if($kooruzly[$i]{afun} =~ m/^(pnct|coord)/ && $i<$#kooruzly && $kooruzly[$i+1]{afun} =~ m/^conj/)
                        {
                            pripojit($kooruzly[$i], $kooruzly[$i+1]);
                            $kooruzly[$i]{_koordinace_zpracovana} = 1;
                        }
                        # Člen koordinace pověsit na předcházející člen koordinace.
                        elsif($kooruzly[$i]{afun} =~ m/^conj/)
                        {
                            pripojit($kooruzly[$i], $predch);
                            $predch = $kooruzly[$i];
                            $kooruzly[$i]{_koordinace_zpracovana} = 1;
                        }
                    }
                    # Zapamatovat si, že tato koordinace už je zpracovaná.
                    # Jinak by se mohlo stát, že u dalšího uzlu se spustí zpracování nanovo.
                    $koren->{_koordinace_zpracovana} = 1;
                    $zmena = 1;
                }
            }
        }
    }
    # Vytisknout větu.
    csts::psat_vetu($veta);
}



#------------------------------------------------------------------------------
# Odpojí uzel od rodiče.
#------------------------------------------------------------------------------
sub odpojit
{
    my $uzel = shift;
    # Vyřadit uzel ze seznamu dětí rodiče.
    my $rodic = $uzel->{parent};
    for(my $i = 0; $i<=$#{$rodic->{children}}; $i++)
    {
        if($rodic->{children}[$i]==$uzel)
        {
            splice(@{$rodic->{children}}, $i, 1);
            last;
        }
    }
    # Odstranit číselný i přímý odkaz na rodiče.
    $uzel->{parentord} = "";
    $uzel->{parent} = "";
}



#------------------------------------------------------------------------------
# Připojí uzel k novému rodiči.
#------------------------------------------------------------------------------
sub pripojit
{
    my $uzel = shift;
    my $rodic = shift;
    # Nejdříve odpojit uzel od dosavadního rodiče.
    odpojit($uzel);
    # Přidat uzel do seznamu dětí rodiče.
    push(@{$rodic->{children}}, $uzel);
    # Přidat číselný i přímý odkaz na rodiče.
    $uzel->{parentord} = $rodic->{ord};
    $uzel->{parent} = $rodic;
    # Označit uzel jako převěšený.
    $uzel->{afun} .= " (převěšeno)" unless($uzel->{afun} =~ m/převěšeno/);
}
