#!/usr/bin/perl
# Normalizuje švédské stromy a převede v nich Hajičovy morfologické značky do sady Penn Treebanku.
# Současně delexikalizuje terminály. Slova nahradí Hajičovými morfologickými značkami.
# (c) 2006 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use csts1;
use tagset::sv::mamba;
use tagset::sv::hajic;
use tagset::en::penn;
use tagset::sv::svdahybrid;



csts::cist(\&zpracovat_vetu);



#------------------------------------------------------------------------------
# Zpracuje strom.
#------------------------------------------------------------------------------
sub zpracovat_vetu
{
    my $veta = shift;
    # Dekódovat morfologické značky do sad vlastností.
    foreach my $uzel (@{$veta})
    {
        my $tag = $uzel->{tag};
        $uzel->{features} = tagset::sv::hajic::decode($tag);
        $uzel->{tag} = tagset::en::penn::encode($uzel->{features});
        $uzel->{form} = tagset::sv::svdahybrid::encode($uzel->{features});
    }
    # Projít strom shora dolů, hledat neobvyklá zavěšení a opravovat je.
    zkontrolovat_podstrom($veta->[0]);
    # Vytisknout větu.
    csts::psat_vetu($veta);
}



#------------------------------------------------------------------------------
# Zkontroluje podstrom určitého uzlu.
#------------------------------------------------------------------------------
sub zkontrolovat_podstrom
{
    my $uzel = shift;
    # Zkontrolovat podstromy všech dětí.
    foreach my $dite (@{$uzel->{children}})
    {
        zkontrolovat_podstrom($dite);
    }
    # Zkontrolovat vztahy mezi mnou a mými dětmi.
    zkontrolovat_vedlejsi_vetu($uzel);
}



#------------------------------------------------------------------------------
# Zkontroluje, že podřadící spojka uvozující vedlejší větu je kořenem vedlejší
# věty a nevisí na jejím hlavním slovesu.
#------------------------------------------------------------------------------
sub zkontrolovat_vedlejsi_vetu
{
    my $uzel = shift;
    # Vedlejší věta se kontroluje z pozice jejího původního chybného kořene.
    # Takže není-li tohle slovesný uzel, nezajímá nás.
    return unless($uzel->{features}{pos} eq "verb");
    # Druhá podmínka: Mezi levými dětmi uzlu je podřadící spojka.
    # Pro jistotu seřadit děti uzlu podle slovosledu.
    my @deti = sort{$a->{ord} <=> $b->{ord}}(@{$uzel->{children}});
    # Hledat spojku.
    my $spojka;
    for(my $i = 0; $i<=$#deti; $i++)
    {
        # Zajímají nás jen děti nalevo od slovesa.
        last if($deti[$i]{ord} > $uzel->{ord});
        # Je tohle podřadící spojka?
        if($deti[$i]{features}{pos} eq "conj" && $deti[$i]{features}{subpos} eq "sub")
        {
            $spojka = $deti[$i];
            last;
        }
    }
    # Jestliže jsme spojku nenašli, ukončit kontrolu.
    return unless($spojka);
    # Předpokládáme také, že spojka je zatím bezdětná.
    return if(scalar(@{$spojka->{children}}));
    # Zavěsit sloveso pod spojku.
    pripojit($spojka, $uzel->{parent});
    pripojit($uzel, $spojka);
}



#------------------------------------------------------------------------------
# Odpojí uzel od rodiče.
#------------------------------------------------------------------------------
sub odpojit
{
    my $uzel = shift;
    # Vyřadit uzel ze seznamu dětí rodiče.
    my $rodic = $uzel->{parent};
    for(my $i = 0; $i<=$#{$rodic->{children}}; $i++)
    {
        if($rodic->{children}[$i]==$uzel)
        {
            splice(@{$rodic->{children}}, $i, 1);
            last;
        }
    }
    # Odstranit číselný i přímý odkaz na rodiče.
    $uzel->{parentord} = "";
    $uzel->{parent} = "";
}



#------------------------------------------------------------------------------
# Připojí uzel k novému rodiči.
#------------------------------------------------------------------------------
sub pripojit
{
    my $uzel = shift;
    my $rodic = shift;
    # Nejdříve odpojit uzel od dosavadního rodiče.
    odpojit($uzel);
    # Přidat uzel do seznamu dětí rodiče.
    push(@{$rodic->{children}}, $uzel);
    # Přidat číselný i přímý odkaz na rodiče.
    $uzel->{parentord} = $rodic->{ord};
    $uzel->{parent} = $rodic;
    # Označit uzel jako převěšený.
    $uzel->{afun} .= " (převěšeno)" unless($uzel->{afun} =~ m/převěšeno/);
}
