#!/usr/bin/perl
# Restuffs Penn trees with the original tokens, regardless what there is now.
# In case of sentence length discrepancy, number of tokens in parser output is kept.
# (c) 2006 Dan Zeman <zeman@ufal.mff.cuni.cz>
# License: GNU GPL

sub usage
{
    print STDERR ("Usage: restuff.pl -s source.txt < input.penn > output.penn\n");
    print STDERR ("  source:   tokenized source text\n");
    print STDERR ("  input:    STDIN or arguments; Penn style trees, one per line\n");
    print STDERR ("  output:   input trees restuffed with original source tokens\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use Getopt::Long;

# Get options.
GetOptions('source=s' => \$source);
if($source eq "")
{
    usage();
    die("Missing source file name.\n");
}
# Read input trees.
open(SOURCE, $source) or die("Cannot read from $source: $!\n");
while(<>)
{
    # Strip the line break.
    s/\r?\n$//;
    # Count left round brackets for debugging purposes.
    my $lrb = $_;
    $lrb =~ s/[^\(]//g;
    my $nlrb = length($lrb);
    # Count right round brackets for debugging purposes.
    my $rrb = $_;
    $rrb =~ s/[^\)]//g;
    my $nrrb = length($rrb);
    # Split terminals and right brackets: insert spaces before right brackets.
    s/\)/ \)/g;
    # Split tree into sequence of brackets and terminals.
    # (Note: nonterminals are stuck to left brackets.)
    my @treeparts = split(/\s+/, $_);
    # Get original text line.
    my $orig = <SOURCE>;
    $orig =~ s/\r?\n$//;
    my @tokens = split(/\s+/, $orig);
    # Replace terminals by original tokens.
    @treeparts = map
    {
        my $result;
        if(m/^[\(\)]/)
        {
            $result = $_;
        }
        else
        {
            my $token = shift(@tokens);
            # Beware of round brackets as original tokens!
            $token =~ s/\(/-LRB-/g;
            $token =~ s/\)/-RRB-/g;
            if($token ne "")
            {
                $result = $token;
            }
            else
            {
                $result = $_;
            }
        }
        $result;
    }
    (@treeparts);
    # Remove spaces before right brackets.
    my $tree = join(" ", @treeparts);
    $tree =~ s/\s+\)/\)/g;
    # Count left round brackets for debugging purposes.
    my $lrb = $tree;
    $lrb =~ s/[^\(]//g;
    my $nlrb1 = length($lrb);
    # Count right round brackets for debugging purposes.
    my $rrb = $tree;
    $rrb =~ s/[^\)]//g;
    my $nrrb1 = length($rrb);
    # Make sure we have neither introduced nor removed any round bracket.
    if($nlrb1 != $nlrb || $nrrb1 != $nrrb || $nlrb != $nrrb || $nlrb1 != $nrrb1)
    {
        print STDERR ("number of left round brackets:  before $nlrb, after $nlrb1\n");
        print STDERR ("number of right round brackets: before $nrrb, after $nrrb1\n");
        die;
    }
    # Print tree.
    print("$tree\n");
}
close(SOURCE);
