#!/usr/bin/perl
# Nechá několik parserů váženě hlasovat o výsledku analýzy.
# Čte N vstupů ve formátu CoNLL 2006, na STDOUT píše výsledek v tomtéž formátu.
# Copyright © 2009 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

sub usage
{
    print STDERR ("Usage: vote.pl [weight1] out1.conll [weight2] out2.conll [...] > out.conll\n");
    print STDERR ("       Weights are non-negative integers.\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");

# Jestliže je nultý argument číselný, budeme předpokládat, že všechny sudé argumenty jsou váhy.
# Jinak předpokládáme, že argumenty jsou pouze cesty k souborům a všechny soubory mají váhu 1.
if($ARGV[0] =~ m/^\d+$/)
{
    for(my $i = 0; $i<=$#ARGV; $i += 2)
    {
        $w[$i/2] = $ARGV[$i];
        $o[$i/2] = $ARGV[$i+1];
    }
}
else
{
    @o = @ARGV;
    @w = map {1} (0..$#ARGV);
}
print STDERR ("Weights = ", join(' ', @w), "\n");
print STDERR ("Outputs = ", join(' ', @o), "\n");
# Otevřít všechny vstupní soubory (výstupy parserů).
for(my $i = 0; $i<=$#o; $i++)
{
    open($oh[$i], $o[$i]) or die("Cannot read from $o[$i]: $!\n");
}
# Číst vstupy, dokud žádný z nich neskončil.
$i_veta = 0;
while(!grep {eof($_)} (@oh))
{
    $i_veta++;
    print STDERR ("Reading sentence $i_veta...\n");
    # Pole alternativních anotací věty.
    my @alt;
    for(my $i = 0; $i<=$#oh; $i++)
    {
        # Chceme, aby nultý uzel každé věty byl kořen, ale ten na vstupu vůbec nenajdeme, musíme si ho vytvořit.
        push(@{$alt[$i]}, {'id' => 0});
        # Přístup do pole se nedá použít uvnitř skobiček, tj. <$oh[$i]> nejde, ale <$h> jde.
        my $h = $oh[$i];
        print STDERR ("  Reading from $h...\n");
        while(1)
        {
            my $radek = <$h>;
            print STDERR ("    $radek");
            # Odstranit znak zalomení řádku.
            $radek =~ s/\r?\n$//;
            # Jestliže jsme přečetli prázdný řádek, věta skončila.
            last if($radek =~ m/^\s*$/);
            my $uzel = radek_na_hash($radek);
            push(@{$alt[$i]}, $uzel);
        }
    }
    # Zkontrolovat, že všechny načtené věty mají stejný počet slov.
    if(grep {scalar(@{$_}) != scalar(@{$alt[0]})} (@alt))
    {
        die("For every sentence, all parsers must output the same number of words.\n");
    }
    # Slepit alternativní rozbory věty do jediného pole hashů.
    # A rovnou už sčítat hlasy.
    my @veta = @{$alt[0]};
    for(my $i = 0; $i<=$#veta; $i++)
    {
        my @navrhy = map {my %h = ('head' => $_, 'deprel' => '_', 'hlasu' => 0); \%h} (0..$#veta);
        for(my $j = 0; $j<=$#alt; $j++)
        {
            $veta[$i]{altheads}[$j] = $alt[$j][$i]{head};
            $veta[$i]{altdeprels}[$j] = $alt[$j][$i]{deprel};
            # Dokud jsme @navrhy nepřetřídili, můžeme se spolehnout na to, že $navrhy[$k]{head} == $k.
            $navrhy[$alt[$j][$i]{head}]{hlasu} += $w[$j];
            # Syntaktickou značku můžeme přiřadit několika způsoby.
            # - Můžeme o ní hlasovat zvlášť, a to buď nezávisle na tom, kam jsme uzel zavěsili, nebo tuto informaci nějak zohlednit.
            # - Můžeme vzít značku, kterou přiřadil parser, jehož zavěšení vyhrálo. To jsem zkusil, ale zhoršilo to úspěšnost značkování.
            #   (Nebylo navíc ošetřeno, když blbý a dobrý parser chtěli stejné vítězné zavěšení, ale každý mu dával jinou značku.)
            # - Nebo můžeme vzít natvrdo značku prostředního parseru ([1]), což je u mě teď MST parser, který umí značky nejlíp.
            #$navrhy[$alt[$j][$i]{head}]{deprel} = $alt[$j][$i]{deprel};
            $navrhy[$alt[$j][$i]{head}]{deprel} = $alt[1][$i]{deprel}; ###!!! viz vyse
            # Kvůli snadnější pozdější manipulaci si ke každému návrhu uložíme také index závislého uzlu,
            # takže návrh vlastně reprezentuje kompletní informaci o závislosti mezi dvěma uzly.
            $navrhy[$alt[$j][$i]{head}]{dep} = $i;
        }
        # Seřadit návrhy sestupně podle počtu hlasů.
        @navrhy = sort {$b->{hlasu} <=> $a->{hlasu}} (@navrhy);
        # Uložit si seřazené návrhy u uzlu.
        $veta[$i]{navrhy} = \@navrhy;
    }
    # Vybudovat strom z návrhů, které dostaly nejvíce hlasů.
    # Na začátku se za "hotový" uzel považuje pouze kořen.
    $veta[0]{hotovo} = 1;
    # Projít uzly a připojit všechny, které preferují rodiče, který už ve stromu je (= je hotov).
    my $hotovo = 0;
    while(!$hotovo)
    {
        my $zmena = 0;
        $hotovo = 1;
        foreach my $uzel (@veta)
        {
            if(!$uzel->{hotovo})
            {
                if($veta[$uzel->{navrhy}[0]{head}]{hotovo})
                {
                    $uzel->{vitez} = $uzel->{navrhy}[0];
                    $uzel->{hotovo} = 1;
                    # Zapamatovat si, že v tomto průchodu se podařilo zavěsit alespoň jeden nový uzel.
                    $zmena = 1;
                }
                else
                {
                    # Zapamatovat si, že ještě zbývá nejméně jeden uzel, který se nepodařilo zavěsit.
                    $hotovo = 0;
                }
            }
        }
        # Pokud zbývají uzly, které nejde připojit ke stromu, znamená to, že na sebe odkazují cyklicky.
        # V tom případě najít uzel, jehož odpojení z cyklu napáchá nejmenší škody.
        if(!$hotovo && !$zmena)
        {
            # Pole návrhů na rodiče u všech uzlů jsou vlastně poli hran (nesou informaci o řídícím i závislém uzlu).
            # Spojíme-li tato pole od všech uzlů do jednoho, máme pole možných hran.
            my @hrany;
            foreach my $uzel (@veta)
            {
                unless($uzel->{hotovo})
                {
                    push(@hrany, @{$uzel->{navrhy}});
                }
            }
            # Všechny hrany mají závislý uzel nehotový.
            # Seřadit hrany tak, aby:
            # - hrany s hotovými řídícími uzly byly na začátku
            # - mezi nimi aby se dopředu dostaly ty, které mají nejvyšší počet hlasů
            @hrany = sort
            {
                my $vysledek = $veta[$b->{head}]{hotovo} <=> $veta[$a->{head}]{hotovo};
                unless($vysledek)
                {
                    $vysledek = $b->{hlasu} <=> $a->{hlasu};
                }
                return $vysledek;
            }
            (@hrany);
            # Zkontrolovat, že vítězná hrana má řídící uzel hotový. Pokud ne, něco není v pořádku.
            die if(!$veta[$hrany[0]{head}]{hotovo});
            # Připojit vítěznou hranu ke stromu.
            $veta[$hrany[0]{dep}]{vitez} = $hrany[0];
            $veta[$hrany[0]{dep}]{hotovo} = 1;
        }
    }
    # Teď už víme, jak má vypadat vítězný strom. Vypsat větu na STDOUT a přejít k další větě.
    for(my $i = 1; $i<=$#veta; $i++)
    {
        # Upravený výstup vypsat.
        my @cells =
        (
            $veta[$i]{id},
            $veta[$i]{form},
            $veta[$i]{lemma},
            $veta[$i]{cpostag},
            $veta[$i]{postag},
            $veta[$i]{feats},
            $veta[$i]{vitez}{head},
            $veta[$i]{vitez}{deprel},
            $veta[$i]{phead},
            $veta[$i]{pdeprel}
        );
        print(join("\t", @cells), "\n");
    }
    print("\n");
}
# Zavřít všechny vstupní soubory.
foreach my $h (@oh)
{
    close($h);
}



#------------------------------------------------------------------------------
# Převede řádek formátu CoNLL 2006 na hash popisující slovo.
#------------------------------------------------------------------------------
sub radek_na_hash
{
    # Předpokládáme, že řádek už neobsahuje znak konce řádku.
    my $radek = shift;
    my @nazvy = qw(id form lemma cpostag postag feats head deprel phead pdeprel);
    my @hodnoty = split(/\s+/, $radek);
    my %uzel;
    for(my $i = 0; $i<=$#nazvy; $i++)
    {
        $uzel{$nazvy[$i]} = $hodnoty[$i];
    }
    return \%uzel;
}
