#!/usr/bin/perl
# Obálka na proces v Linuxu, která zaznamená, jak dlouho proces běžel a kolik potřeboval paměti.
# Copyright © 2009 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

# Návrh budoucího ovládání:
# První argument může být volba (začínat pomlčkou), pak je druhý argument argumentem volby a teprve další argumenty jsou volání měřeného programu.
# Bez volby: psát na STDERR, řádky začínat ZMVYK:
# Volba -e xyz: psát na STDERR, řádky začínat ZMVYK xyz:
# Volba -f x.log: psát do souboru x.log, nezačínat nijak
#
# Na začátku vypsat značku procesoru, jeho taktovací frekvenci a velikost operační paměti.
# V průběhu vypisovat každou sekundu aktuální obsazenost paměti. To se může hodit pro ladění, když program předčasně skončí kvůli nedostatku paměti.
# Na konci vypsat maximální obsazenou paměť (RSIZE, ne virtuální), celkovou dobu trvání programu (včetně času, kdy nic nedělal a třeba čekal na diskové operace) a dobu strávenou na procesoru.

# Tento skript neumí procházet strom všech procesů-potomků, takže dokáže změřit
# jen jeden proces. Pokud se tento proces dále rozdělí a hlavní část práce svěří
# svým potomkům, nezjistíme to. Pokud např. měříme make, hlavní nápor na systémové
# prostředky nejspíš způsobí programy, které make spustí, ale my změříme pouze
# nároky maku samotného.

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");

$procesor = zjistit_procesor();
print STDERR ("ZMVYK: Typ procesoru: $procesor->{'model name'}\n");
print STDERR ("ZMVYK: Frekvence:     $procesor->{'cpu MHz'}\n");
print STDERR ("ZMVYK: Rychlost MIPS: $procesor->{'bogomips'}\n");
$pamet = zjistit_pamet();
print STDERR ("ZMVYK: Celková paměť: $pamet->{'MemTotal'}\n");
print STDERR ("ZMVYK: Volná paměť:   $pamet->{'MemFree'}\n");
$starttime = time();
print STDERR ("ZMVYK: Výpočet začal v ".cas($starttime).".\n");
$prikaz = join(' ', @ARGV);
# Rozvětvit proces na dva. Rodič bude měřit pomocí linuxových nástrojů práci dítěte.
$child_pid = fork();
if($child_pid==0)
{
    # Já jsem dítě, mám tedy pustit proces, jehož výkon měříme.
    # Rodič má moje PID a bude měřit mě. Nemohu proto měřený proces zavolat, musím se v něj převtělit.
    exec($prikaz) or die("Nelze spustit $prikaz: $!\n");
}
else
{
    # Já jsem rodič, mám tedy měřit výkon procesu dítěte.
    $prikaz = "ps -o \"rss=\" -p $child_pid";
    while(1)
    {
        $memory = `$prikaz`;
        $memory =~ s/^\s+//;
        $memory =~ s/\r?\n$//;
        if($memory)
        {
            print STDERR ("ZMVYK: Obsazená operační paměť: $memory kB\n");
        }
        else
        {
            print STDERR ("ZMVYK: Proces uvolnil operační paměť, takže asi skončil.\n");
            $stoptime = time();
            print STDERR ("ZMVYK: Výpočet skončil v ".cas($stoptime).".\n");
            my $trvani = zjistit_delku_trvani_programu($starttime, $stoptime);
            print STDERR ("ZMVYK: Proces běžel $trvani hodin.\n");
            exit();
        }
        sleep(1);
    }
}



#------------------------------------------------------------------------------
# Vrátí aktuální čas jako řetězec s položkami oddělenými dvojtečkou. Délka
# řetězce je vždy stejná (8 znaků), což lze využít při sloupcovém formátování.
#------------------------------------------------------------------------------
sub cas
{
    my($h, $m, $s);
    ($s, $m, $h) = localtime(time());
    return sprintf("%02d:%02d:%02d", $h, $m, $s);
}



#------------------------------------------------------------------------------
# Vypíše dobu, po kterou program běžel. K tomu potřebuje dostat časové otisky
# začátku a konce.
#------------------------------------------------------------------------------
sub zjistit_delku_trvani_programu
{
    my $starttime = shift;
    my $stoptime = shift;
    if($soubor eq "")
    {
        $soubor = "konfig";
    }
    my $cas = $stoptime-$starttime;
    my $hod = int($cas/3600);
    my $min = int(($cas%3600)/60);
    my $sek = $cas%60;
    return sprintf("%02d:%02d:%02d", $hod, $min, $sek);
}



#------------------------------------------------------------------------------
# Zjistí typ procesoru, na kterém běžíme. Využívá k tomu soubor, který je
# k dispozici na některých instalacích Linuxu, ale na jiných strojích nebude
# fungovat.
#------------------------------------------------------------------------------
sub zjistit_procesor
{
    my %zaznam;
    open(INF, "/proc/cpuinfo") or return;
    while(<INF>)
    {
        # Záznamy v /proc/cpuinfo se na víceprocesorových strojích opakují pro jednotlivé procesory.
        # Předpokládáme, že všechny procesory na jednom stroji mají stejné parametry.
        s/\r?\n$//;
        if(m/^(.*?)\s*:\s*(.*)$/)
        {
            my $atribut = $1;
            my $hodnota = $2;
            $zaznam{$atribut} = $hodnota;
        }
    }
    close(INF);
    return \%zaznam;
}



#------------------------------------------------------------------------------
# Zjistí velikost operační paměti, kterou má k dispozici stroj, na kterém
# běžíme. Využívá k tomu soubor, který je k dispozici na některých instalacích
# Linuxu, ale na jiných strojích nebude fungovat.
#------------------------------------------------------------------------------
sub zjistit_pamet
{
    my %zaznam;
    open(INF, "/proc/meminfo") or return;
    while(<INF>)
    {
        s/\r?\n$//;
        if(m/^(.*?)\s*:\s*(.*)$/)
        {
            my $atribut = $1;
            my $hodnota = $2;
            $zaznam{$atribut} = $hodnota;
        }
    }
    close(INF);
    return \%zaznam;
}
